/*
 *  JapeParms.java
 *
 */

package at.ofai.gate.japeparms;

import java.io.IOException;
import java.net.MalformedURLException;
import java.util.*;

import gate.*;
import gate.creole.*;
import gate.creole.metadata.*;
import gate.creole.ontology.Ontology;
import gate.util.*;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URLConnection;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.apache.commons.io.FileUtils;
import org.apache.commons.io.IOUtils;
import org.stringtemplate.v4.*;

/** 
 * This class is the implementation of the PR resource JapeParms.
 * The resource wraps a JAPE Transducer and replaces variables in the 
 * JAPE file with the values from a feature map that is specified as
 * init-parameter "variables" before passing the modified JAPE file on
 * to the wrapped JAPE Transducers.
 * <p>
 * The JAPE grammar should follow the syntax needed by the String Template
 * Engine: http://www.antlr.org/wiki/display/ST4/StringTemplate+4+Documentation
 * Variables must be delimited by "$" at the beginning and end, not enclosed
 * between "<" and ">", for example: $var1$ would be the template variable 
 * var1.
 * <p>
 * The variables from the feature map are not only replaced in the JAPE 
 * grammer template but also passed on as the value of feature "$VARS$"
 * for ctx.getPRFeatures().
 * <p>
 * NOTE: this is a preliminary implementation and most of it may change
 * in the future!!
 * 
 * @author Johann Petrak
 * 
 */
@CreoleResource(name = "JapeParms",
        comment = "Add a descriptive comment about this resource")
public class JapeParms  extends AbstractLanguageAnalyser 
    implements ControllerAwarePR, Benchmarkable {

  
  @CreoleParameter(comment="Select a predefined known JAPE Transducer. To specify you own use otherTransducerClassname",defaultValue="JAPE_CLASSIC")
  public void setKnownTransducer(KnownJapeTransducers which) {
    knownJapeTransducer = which;
  }
  public KnownJapeTransducers getKnownTransducer() {
    return knownJapeTransducer;
  }
  private KnownJapeTransducers knownJapeTransducer = KnownJapeTransducers.JAPE_CLASSIC;

  @Optional
  @CreoleParameter(comment="If given, use this classname instead of a known one",
    defaultValue="")
  public void setOtherTransducerClassname(String classname) {
    otherTransducerClassname = classname;
  }
  public String getOtherTransducerClassname() {
    return otherTransducerClassname;
  } 
  private String otherTransducerClassname = "";
  
  @Optional
  @CreoleParameter(comment="Specify where the JAPE Transducer is available if needed")
  public void setPluginDirectoryURL(java.net.URL theURL) {
    pluginDirectoryURL = theURL;
  }
  public java.net.URL getPluginDirectoryURL() { return pluginDirectoryURL; }
  private java.net.URL pluginDirectoryURL;
  
  
  // replicate the other common JAPE parameters here.
  @Optional
  @CreoleParameter(
    comment = "Class names that implement gate.jape.constraint.AnnotationAccessor."
  )
  public void setAnnotationAccessors(List<String> annotationAccessors) {
    this.annotationAccessors = annotationAccessors;
  }
  public List<String> getAnnotationAccessors() {
    return annotationAccessors;
  }
  private List<String> annotationAccessors;

  public void setBenchmarkId(String benchmarkId) {
    if(theTransducerPR != null && theTransducerPR instanceof Benchmarkable) {
      ((Benchmarkable)theTransducerPR).setBenchmarkId(benchmarkId);
    }
  }
  public String getBenchmarkId() { 
    if(theTransducerPR != null && theTransducerPR instanceof Benchmarkable) {
      return ((Benchmarkable)theTransducerPR).getBenchmarkId();
    } else {
      return null;
    }
  }
  
  
  @CreoleParameter(
    comment = "The encoding used for reading the grammar",
    defaultValue = "UTF-8"
  )
  public void setEncoding(String newEncoding) {
    encoding = newEncoding;
  }
  public String getEncoding() {
    return encoding;
  }
  private String encoding = "UTF-8";
  
  @CreoleParameter(
    comment = "The URL to the grammar file. Variables should be enclosed between $ chars.",
    suffixes = "jape"
  )
  public void setGrammarURL(java.net.URL newGrammarURL) {
    grammarURL = newGrammarURL;
  }
  public java.net.URL getGrammarURL() {
    return grammarURL;
  }
  private java.net.URL grammarURL;

  @Optional
  @CreoleParameter(
    comment = "The file: URL of a directory where to store the generated file. The default is either a known tmp dir or the same dir the grammar file is in."
  )
  public void setTempDirectoryURL(java.net.URL url) {
    tempDirectoryURL = url;
  }
  public java.net.URL getTempDirectoryURL() {
    return tempDirectoryURL;
  }
  private java.net.URL tempDirectoryURL;
  
  @CreoleParameter(
    comment = "Variables to use and substitute in the JAPE grammer template. Do not include the enclosing $s"
  )
  public void setVariables(FeatureMap vars) {
    variables = vars;
  }
  public FeatureMap getVariables() {
    return variables;
  }
  private FeatureMap variables;
  
  
  
  
  @RunTime
  @Optional
  @CreoleParameter(
    comment = "The annotation set to be used as input for the transducer",
    defaultValue = ""
  )
  public void setInputASName(String newInputASName) {
    inputASName = newInputASName;
  }
  public String getInputASName() {
    return inputASName;
  }
  private String inputASName = "";

  @RunTime
  @Optional
  @CreoleParameter(
    comment = "The annotation set to be used as output for the transducer",
    defaultValue = ""
  )
  public void setOutputASName(String newOutputASName) {
    outputASName = newOutputASName;
  }
  public String getOutputASName() {
    return outputASName;
  }
  private String outputASName = "";
  
  @RunTime
  @Optional
  @CreoleParameter(
    comment = "The ontology to be used by this transducer"
  )
  public void setOntology(Ontology ontology) {
    this.ontology = ontology;
  }
  public Ontology getOntology() {
    return ontology;
  }
  private Ontology ontology;

  @Optional
  @CreoleParameter(
    comment = "Class names that implement gate.jape.constraint.ConstraintPredicate."
  )
  public void setOperators(List<String> operators) {
    this.operators = operators;
  }
  public List<String> getOperators() {
    return operators;
  }
  private List<String> operators;
  
  protected LanguageAnalyser theTransducerPR;
  protected File generatedGrammarFile;
  protected java.net.URL generatedFileURL = null;
  
  @Override
  public Resource init() throws ResourceInstantiationException {  
    // Check that required parameters are given
    if(variables == null) {
      throw new ResourceInstantiationException("Parameter variables must be specified");
    }
    // try to create the requested transducer PR
    String usedTransducerClassname = "";
    if(otherTransducerClassname == null || otherTransducerClassname.isEmpty()) {
      usedTransducerClassname = knownJapeTransducer.getClassName();
      String directoryName = knownJapeTransducer.getPluginDirectoryName();
      java.net.URL usedDirectoryURL = null;
      if(directoryName.isEmpty()) {
        if(getPluginDirectoryURL() == null) {
          throw new ResourceInstantiationException("No known plugin directory for JAPE transducer");
        }
        usedDirectoryURL = getPluginDirectoryURL();
      } else {
        // construct the URL from gate home and the known directory name
        File pluginDir = new File(new File(Gate.getGateHome(),"plugins"),directoryName);
        try {
          usedDirectoryURL = pluginDir.toURI().toURL();
        } catch (MalformedURLException ex) {
          throw new ResourceInstantiationException("Could not convert plugin dir to URL: "+usedDirectoryURL,ex);
        }
      }
      try {
        Gate.getCreoleRegister().registerDirectories(usedDirectoryURL);
      } catch (GateException ex) {
        throw new ResourceInstantiationException("Could not register plugin directory "+usedDirectoryURL,ex);
      }
    } else {
      usedTransducerClassname = otherTransducerClassname;
    }
    
    FeatureMap parms = Factory.newFeatureMap();
    FeatureMap feats = Factory.newFeatureMap();
    feats.putAll(features);
    Gate.setHiddenAttribute(feats, true);
    feats.put("$VARS$", getVariables());
    String prname = this.getName() + "_wrapped";
    // copy over the init parms
    parms.put("encoding", getEncoding());
    // this method updates generatedFileURL, if necessary
    regenerateInputFile(grammarURL,tempDirectoryURL, variables);
    parms.put("grammarURL",generatedFileURL);
    theTransducerPR = 
      (LanguageAnalyser)Factory.createResource(usedTransducerClassname, parms, feats, prname);
    // hide the wrapped PR
    return this;
  }
  
  @Override
  public void reInit() throws ResourceInstantiationException {
    regenerateInputFile(grammarURL, tempDirectoryURL, variables);
    theTransducerPR.reInit();
  }
  
  
  // This (re-)creates a temporary file from the original JAPE file using the
  // stringtemplate library 
  // If the dirURL is null, we try to find some system directory for temporary files
  // If all else fails, we use the current directory.
  protected void regenerateInputFile(java.net.URL originalURL, 
    java.net.URL dirURL, FeatureMap variables) throws ResourceInstantiationException {
    if(generatedFileURL == null) {
      File tmpDir = new File(System.getProperty("gate.user.filechooser.defaultdir"));
      if(dirURL != null) {
        tmpDir = Files.fileFromURL(dirURL);
      } else {
        String tmplocation = System.getProperty("java.io.tmpdir");
        if(tmplocation != null) {
          tmpDir = new File(tmplocation);
        }
      }
      File grammarfile = gate.util.Files.fileFromURL(grammarURL);
      String fileName = "JapeParms_" + getName() + "_" + grammarfile.getName() + "_" + Gate.genSym() + ".jape";
      //System.out.println("Generated file name: "+fileName);
      try {
        generatedFileURL = new File(tmpDir,fileName).toURI().toURL();
      } catch (MalformedURLException ex) {
        throw new ResourceInstantiationException("Cannot create generated file URL for "+tmpDir+" / "+fileName);
      }
    }
    String jape = "";
    try {
      jape = readURL2String(originalURL,encoding);
    } catch (IOException ex) {
      throw new ResourceInstantiationException("Could not read grammar file "+originalURL,ex);
    }
    //System.out.println("Got jape file: \n"+jape);
    ST template = new ST(jape,'$','$');
    //System.out.println("Created template");
    for(Object key : variables.keySet()) {
      if(key instanceof String) {
        template.add((String)key, variables.get(key));
      }
    }
    String result = template.render();
    //System.out.println("Rendered template: \n"+result);
    File generatedFile = Files.fileFromURL(generatedFileURL);
    System.out.println("Generated grammar file URL: "+generatedFileURL);
    try {
      FileUtils.writeStringToFile(generatedFile, result, encoding);
    } catch (IOException ex) {
      throw new ResourceInstantiationException("Could not write to file "+generatedFile,ex);
    }
  }
  
  protected String readURL2String(java.net.URL url, String encoding) throws IOException {
    InputStream is = url.openStream();
    String result = IOUtils.toString(is, encoding);
    is.close();
    return result;
  }
  
  
  @Override
  public void execute() throws ExecutionException {
    // update runtime parms, then 
    try {
      theTransducerPR.setParameterValue("inputASName", inputASName);
      theTransducerPR.setParameterValue("outputASName", outputASName);
      theTransducerPR.setParameterValue("ontology", ontology);
    } catch (ResourceInstantiationException ex) {
      throw new GateRuntimeException("Could not set runtime parameter value",ex);
    }
    theTransducerPR.setCorpus(corpus);
    theTransducerPR.setDocument(document);
    theTransducerPR.execute();
  }
  
  public synchronized void interrupt() {
    theTransducerPR.interrupt();
  }

  @Override
  public void cleanup() {
    // remove the wrapped Transducer
    Factory.deleteResource(theTransducerPR);
    // remove the generated file
    if(generatedFileURL != null) {
      File generatedFile = gate.util.Files.fileFromURL(generatedFileURL);
      generatedFile.delete(); // silently ignore unsuccessful delete (false returned)
    }
  }
  
  public void controllerExecutionStarted(Controller c) throws ExecutionException {
    if(theTransducerPR instanceof ControllerAwarePR) {
      ((ControllerAwarePR)theTransducerPR).controllerExecutionStarted(c);
    }
  }
  
  public void controllerExecutionFinished(Controller c)
    throws ExecutionException {
    if(theTransducerPR instanceof ControllerAwarePR) {
      ((ControllerAwarePR)theTransducerPR).controllerExecutionFinished(c);
    }
  }
  
  public void controllerExecutionAborted(Controller c, Throwable t)
    throws ExecutionException {
    if(theTransducerPR instanceof ControllerAwarePR) {
      ((ControllerAwarePR)theTransducerPR).controllerExecutionAborted(c, t);
    }
  }
  
  
} // class JapeParms
